# Copyright 2024 The bayeux Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Imports from submodules."""
# pylint: disable=g-importing-member
# pylint: disable=g-import-not-at-top
# pylint: disable=g-bad-import-order
import importlib

# TFP-on-JAX always installed
from bayeux._src.mcmc.tfp import HMC as HMC_TFP
from bayeux._src.mcmc.tfp import NUTS as NUTS_TFP
from bayeux._src.mcmc.tfp import SnaperHMC as SNAPER_HMC_TFP
__all__ = ["HMC_TFP", "NUTS_TFP", "SNAPER_HMC_TFP"]

if importlib.util.find_spec("blackjax") is not None:
  from bayeux._src.mcmc.blackjax import CheesHMC as CheesHMCblackjax
  from bayeux._src.mcmc.blackjax import HMC as HMCblackjax
  from bayeux._src.mcmc.blackjax import HMCPathfinder as HMC_Pathfinder_blackjax
  from bayeux._src.mcmc.blackjax import MeadsHMC as MeadsHMCblackjax
  from bayeux._src.mcmc.blackjax import NUTS as NUTSblackjax
  from bayeux._src.mcmc.blackjax import NUTSPathfinder as NUTS_Pathfinder_blackjax
  __all__.extend(["HMCblackjax", "CheesHMCblackjax", "MeadsHMCblackjax",
                  "NUTSblackjax", "HMC_Pathfinder_blackjax",
                  "NUTS_Pathfinder_blackjax"])

if importlib.util.find_spec("flowMC") is not None:
  from bayeux._src.mcmc.flowmc import MaskedCouplingRQSplineHMC as MaskedCouplingRQSplineHMCflowmc
  from bayeux._src.mcmc.flowmc import MaskedCouplingRQSplineMALA as MaskedCouplingRQSplineMALAflowmc
  from bayeux._src.mcmc.flowmc import RealNVPHMC as RealNVPHMCflowmc
  from bayeux._src.mcmc.flowmc import RealNVPMALA as RealNVPMALAflowmc

  __all__.extend([
      "MaskedCouplingRQSplineHMCflowmc",
      "MaskedCouplingRQSplineMALAflowmc",
      "RealNVPHMCflowmc",
      "RealNVPMALAflowmc"])

if importlib.util.find_spec("numpyro") is not None:
  from bayeux._src.mcmc.numpyro import HMC as HMCnumpyro
  from bayeux._src.mcmc.numpyro import NUTS as NUTSnumpyro

  __all__.extend(["HMCnumpyro", "NUTSnumpyro"])



# Copyright 2024 The bayeux Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""JAXopt specific code."""
import functools

from bayeux._src.optimize import shared
import jax
import jaxopt


_OPTIMIZER_FNS = {
    "BFGS": jaxopt.BFGS,
    "GradientDescent": jaxopt.GradientDescent,
    "LBFGS": jaxopt.LBFGS,
    "NonlinearCG": jaxopt.NonlinearCG,
}


class _JAXoptOptimizer(shared.Optimizer):
  """Base class for JAXopt samplers."""

  def get_kwargs(self, **kwargs):
    optimizer = _OPTIMIZER_FNS[self.optimizer]
    return {
        optimizer: shared.get_optimizer_kwargs(
            optimizer, kwargs, ignore_required={"fun"}),
        "extra_parameters": shared.get_extra_kwargs(kwargs)}

  def __call__(self, seed, **kwargs):
    kwargs = self.get_kwargs(**kwargs)
    fun, initial_state, apply_transform = self._prep_args(seed, kwargs)

    optimizer_fn = _OPTIMIZER_FNS[self.optimizer]
    jaxopt_cls = optimizer_fn(fun=fun, **kwargs[optimizer_fn])  # pylint: disable=unexpected-keyword-arg
    num_iters = kwargs["extra_parameters"]["num_iters"]
    optimizer = functools.partial(_jaxopt_fit,
                                  jaxopt_cls=jaxopt_cls,
                                  num_iters=num_iters)
    chain_method = kwargs["extra_parameters"]["chain_method"]
    mapped_optimizer = self._map_optimizer(chain_method, optimizer)

    (res, _), loss = mapped_optimizer(initial_state)
    if apply_transform:
      params = self.transform_fn(res[0])
    else:
      params = res[0]
    return shared.OptimizerResults(
        params=params,
        state=res[1],
        loss=loss)


class BFGS(_JAXoptOptimizer):
  name = "jaxopt_bfgs"
  optimizer = "BFGS"


class GradientDescent(_JAXoptOptimizer):
  name = "jaxopt_gradient_descent"
  optimizer = "GradientDescent"


class LBFGS(_JAXoptOptimizer):
  name = "jaxopt_lbfgs"
  optimizer = "LBFGS"


class NonlinearCG(_JAXoptOptimizer):
  name = "jaxopt_nonlinear_cg"
  optimizer = "NonlinearCG"


def _jaxopt_fit(params, jaxopt_cls, num_iters):
  """JAXopt while loop.

  This is needed to run a `while` loop while collecting loss from the loop.

  Args:
    params: Initial position
    jaxopt_cls: A solver class in JAXopt, with methods `init_state`, `update`,
      and the attribute `tol`.
    num_iters: Maximum number of iterations. Loop switches to no-ops after the
      convergence criteria is met.

  Returns:
    (final params, final state), losses.
  """
  state = jaxopt_cls.init_state(params)

  def step(opt_state):
    params, state = jaxopt_cls.update(*opt_state)
    cond = state.error > jaxopt_cls.tol
    return (params, state), cond

  def fun(opt_state, _):
    val, cond = opt_state
    ret = jax.lax.cond(cond, step, lambda x: (x, False), val)
    return ret, ret[0][1].error

  init_val = (params, state)
  return jax.lax.scan(fun, (init_val, True), None, length=num_iters)



# Copyright 2024 The bayeux Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""optimistix specific code."""
from bayeux._src.optimize import shared
import optimistix


class _OptimistixOptimizer(shared.Optimizer):
  """Base class for optimistix optimizers."""

  def get_kwargs(self, **kwargs):
    kwargs = self.default_kwargs() | kwargs
    solver = getattr(optimistix, self.optimizer)
    minimise_kwargs = shared.get_optimizer_kwargs(
        optimistix.minimise, kwargs, ignore_required={"y0", "solver", "fn"})
    for k in minimise_kwargs:
      if k in kwargs:
        minimise_kwargs[k] = kwargs[k]
    extra_parameters = shared.get_extra_kwargs(kwargs)
    _ = extra_parameters.pop("num_iters")
    return {solver: shared.get_optimizer_kwargs(solver, kwargs),
            optimistix.minimise: minimise_kwargs,
            "extra_parameters": extra_parameters}

  def default_kwargs(self) -> dict[str, float]:
    return {"rtol": 1e-5, "atol": 1e-5}

  def _prep_args(self, seed, kwargs):
    fun, initial_state, apply_transform = super()._prep_args(seed, kwargs)
    def f(x, _):
      return fun(x)
    return f, initial_state, apply_transform

  def __call__(self, seed, **kwargs):
    kwargs = self.get_kwargs(**kwargs)
    fun, initial_state, apply_transform = self._prep_args(seed, kwargs)

    solver_fn = getattr(optimistix, self.optimizer)
    def run(x0):
      solver = solver_fn(**kwargs[solver_fn])
      return optimistix.minimise(
          fn=fun,
          solver=solver,
          y0=x0,
          **kwargs[optimistix.minimise]).value
    chain_method = kwargs["extra_parameters"]["chain_method"]
    mapped_run = self._map_optimizer(chain_method, run)
    ret = mapped_run(initial_state)
    if apply_transform:
      return shared.OptimizerResults(
          params=self.transform_fn(ret), state=None, loss=None)
    else:
      return shared.OptimizerResults(ret, state=None, loss=None)


class BFGS(_OptimistixOptimizer):
  name = "optimistix_bfgs"
  optimizer = "BFGS"


class NelderMead(_OptimistixOptimizer):
  name = "optimistix_nelder_mead"
  optimizer = "NelderMead"


class NonlinearCG(_OptimistixOptimizer):
  name = "optimistix_nonlinear_cg"
  optimizer = "NonlinearCG"



# Copyright 2024 The bayeux Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Initialization strategies."""
from typing import Callable

from bayeux._src.types import Point  # pylint: disable=g-importing-member
import jax
import jax.numpy as jnp


def uniform(
    *,
    test_point: Point,
    inverse_transform_fn: Callable[[Point], Point],
    transform_fn: Callable[[Point], Point],
    num_points: int,
    key: jax.Array
) -> Point:
  """A uniform initialization in unconstrained space between -2 and +2.

  Args:
    test_point: A PyTree that is used to determine the correct shape for
      initialization.
    inverse_transform_fn: A function that pulls the points back to R^n.
    transform_fn: A function that pushes points to the support of the log
      density.
    num_points: A number of points to initialize with.
    key: JAX PRNGkey.

  Returns:
    A batch of `num_points` with the same shape as `test_point`.
  """
  # We only need a shape here, but some bijectors may change the shape, so we
  # take the inverse transform, map over the shape, then transform back.
  untransformed = inverse_transform_fn(test_point)
  treedef = jax.tree_util.tree_structure(untransformed)
  keys = jax.random.split(key, treedef.num_leaves)
  draws = jax.tree_util.tree_map(
      lambda k, d: 4 * jax.random.uniform(  # pylint: disable=g-long-lambda
          key=k, shape=(num_points,) + jnp.shape(d)) - 2,
      jax.tree_util.tree_unflatten(treedef, keys), untransformed)

  return transform_fn(draws)



# Copyright 2024 The bayeux Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for debugging models."""
import dataclasses
import pprint
import random
from typing import Any, Callable, Mapping, Sequence, Type

from bayeux._src.types import Point  # pylint: disable=g-importing-member
import jax
import jax.numpy as jnp


class _Marks:
  """Unicode points for debug messages."""

  def __init__(self):
    self._goods = [
        "\N{CHECK MARK}", "\N{THUMBS UP SIGN}", "\N{DOG}", "\N{RAINBOW}",
        "\N{ROCKET}", "\N{FLEXED BICEPS}", "\N{PARTY POPPER}",
        "\N{PERSON RAISING BOTH HANDS IN CELEBRATION}"
    ]
    self._bads = ["\N{MULTIPLICATION SIGN}", "\N{CROSS MARK}",
                  "\N{THUMBS DOWN SIGN}", "\N{SKULL}"]
    self.engaged = False

  def good(self):
    if self.engaged:
      return random.choice(self._goods)
    else:
      return self._goods[0]

  def bad(self):
    if self.engaged:
      return random.choice(self._bads)
    else:
      return self._bads[0]


FunMode = _Marks()
_GOOD = FunMode.good
_BAD = FunMode.bad


DebugState = Mapping[str, Point]
Kwargs = Mapping[str, Any]
_Base = Any


@dataclasses.dataclass(frozen=True)
class DebugAttr:
  parents: Sequence["DebugAttr"]
  getter: Callable[[_Base, DebugState, Kwargs], DebugState]

  def __call__(self, model, debug_state, kwargs):
    if self not in debug_state:
      _ = [parent(model, debug_state, kwargs) for parent in self.parents]
      debug_state[self] = self.getter(model, debug_state, kwargs)
    return debug_state[self]


Seed = DebugAttr(parents=(), getter=lambda model, debug_state, _: None)

Verbosity = DebugAttr(parents=(), getter=lambda model, debug_state, _: None)

TestPoint = DebugAttr(
    parents=(), getter=lambda model, debug_state, _: model.test_point)

# pylint: disable=g-long-lambda
TestPointLogDensity = DebugAttr(
    parents=(TestPoint,),
    getter=lambda model, debug_state, _: model.log_density(
        debug_state[TestPoint]))

DefaultKwargs = DebugAttr(
    parents=(),
    getter=lambda model, debug_state, kwargs: model.get_kwargs(**kwargs))

InitialState = DebugAttr(
    parents=(Seed, DefaultKwargs),
    getter=lambda model, debug_state, _: model.get_initial_state(
        debug_state[Seed],
        num_chains=_get_num_chains(debug_state[DefaultKwargs])))

InitialStateLogDensity = DebugAttr(
    parents=(InitialState,),
    getter=lambda model, debug_state, _: jax.vmap(model.log_density)
    (debug_state[InitialState]))

UnconstrainedState = DebugAttr(
    parents=(InitialState,),
    getter=lambda model, debug_state, _: model.inverse_transform_fn(debug_state[
        InitialState]))

UnconstrainedLogDensityNoILDJ = DebugAttr(
    parents=(UnconstrainedState,),
    getter=lambda model, debug_state, _: jax.vmap(
        lambda x: model.log_density(model.transform_fn(x))
    )(debug_state[UnconstrainedState]),
)

UnconstrainedLogDensity = DebugAttr(
    parents=(UnconstrainedState,),
    getter=lambda model, debug_state, _: jax.vmap(
        model.constrained_log_density())(debug_state[UnconstrainedState]),
)

InitialGradientsNoILDJ = DebugAttr(
    parents=(UnconstrainedState,),
    getter=lambda model, debug_state, _: jax.vmap(
        jax.grad(model.transformed_negative_log_prob())
    )(debug_state[UnconstrainedState]),
)

InitialGradients = DebugAttr(
    parents=(UnconstrainedState,),
    getter=lambda model, debug_state, _: jax.vmap(
        jax.grad(model.constrained_log_density()))
    (debug_state[UnconstrainedState]),
)
# pylint: enable=g-long-lambda


def make_debug_print(printer):
  """Returns a debug print function."""
  def debug_print(*, msg, verbosity, success=None, end="\n", sep=" ",):
    if verbosity <= 1:
      if success is None:
        return
      msg = ""
      sep = ""
    if verbosity < 2:
      end = " "

    if success is None:
      msg = [msg]
    else:
      if success:
        msg = [_GOOD(), msg]
      else:
        msg = [_BAD(), msg]
    printer(*msg, end=end, sep=sep)
  return debug_print


@dataclasses.dataclass(frozen=True)
class DebugStep:
  """A single step of debugging."""
  description: str
  attrs: Sequence[DebugAttr]
  check_fn: Callable[[_Base, DebugState, Kwargs], bool]
  success_msg: Callable[[DebugState], str]
  failure_msgs: Sequence[Callable[[DebugState], str]] = tuple()
  exceptions: tuple[Type[BaseException], ...] = tuple()

  def _run_with_exceptions(
      self,
      *,
      fn,
      model,
      debug_state,
      verbosity,
      debug_print,
      kwargs,
      catch_exceptions: bool,
  ):
    """Runs a debug step and tries to catch any errors."""
    try:
      ret = fn(model, debug_state, kwargs)
      return ret, True, debug_state
    except self.exceptions as exc:  # pylint: disable=catching-non-exception
      msg = self.failure_msgs[self.exceptions.index(type(exc))]
      debug_print(
          msg=msg,
          verbosity=verbosity - 1,
          success=False)
      return None, False, debug_state
    except Exception as exc:
      debug_print(
          msg=f"Unexpected error!\n{exc!r}",
          verbosity=verbosity - 1,
          success=False)
      if catch_exceptions:
        debug_print(
            msg="Run with `catch_exceptions=False` to raise the full exception",
            verbosity=verbosity - 1)
        return None, False, debug_state
      raise exc

  def __call__(
      self,
      *,
      model,
      debug_state,
      verbosity,
      debug_print,
      kwargs,
      catch_exceptions: bool,
  ):
    debug_print(
        msg=self.description,
        verbosity=verbosity,
        end=" ")
    if any(attr not in debug_state for attr in self.attrs):
      for attr in self.attrs:
        _, ran, debug_state = self._run_with_exceptions(
            fn=attr,
            model=model,
            debug_state=debug_state,
            verbosity=verbosity,
            debug_print=debug_print,
            kwargs=kwargs,
            catch_exceptions=catch_exceptions)
        if not ran:
          return False, debug_state
    ret, ran, debug_state = self._run_with_exceptions(
        fn=self.check_fn,
        model=model,
        debug_state=debug_state,
        verbosity=verbosity,
        debug_print=debug_print,
        kwargs=kwargs,
        catch_exceptions=catch_exceptions)
    success = bool(ran and ret)
    if success:
      msg = self.success_msg(debug_state)
    else:
      msg = self.failure_msgs[-1](debug_state)
    debug_print(
        msg="",
        verbosity=verbosity - 1,
        success=success)
    debug_print(
        msg=msg,
        verbosity=verbosity - 1,
        end="\n")
    return success, debug_state


def _always_succeed(model, debug_state, kwargs):
  del model, debug_state, kwargs
  return True


def _get_shape(pytree):
  return jax.tree_util.tree_map(jnp.shape, pytree)


def pytree_to_msg(pytree, pytree_name, verbosity) -> str:
  if verbosity < 3:
    return ""
  elif verbosity == 3:
    return f"{pytree_name} has shape\n{_get_shape(pytree)!r}"
  elif verbosity > 3:
    return f"{pytree_name} is\n{pytree!r}"
  else:
    return ""

# pylint: disable=g-long-lambda
check_shapes = DebugStep(
    description="Checking test_point shape",
    attrs=(TestPoint,),
    check_fn=_always_succeed,
    success_msg=lambda state: pytree_to_msg(state[TestPoint], "Test point",
                                            state[Verbosity]))


def _check_test_point_log_density(model, debug_state, kwargs):
  del model, kwargs

  scalar_shape = ~bool(jnp.shape(debug_state[TestPointLogDensity]))
  is_real = ~jnp.isnan(debug_state[TestPointLogDensity])
  return scalar_shape and is_real


check_test_point_log_density = DebugStep(
    description="Computing test point log density",
    attrs=(TestPointLogDensity,),
    check_fn=_check_test_point_log_density,
    success_msg=lambda state:
    f"Test point has log density\n{state[TestPointLogDensity]!r}",
    failure_msgs=[
        lambda state:
        f"Test point has log density\n{state[TestPointLogDensity]!r}"
    ])

check_kwargs = DebugStep(
    description="Loading keyword arguments...",
    attrs=(DefaultKwargs,),
    check_fn=_always_succeed,
    success_msg=lambda state:
    f"Keyword arguments are\n{pprint.pformat(state[DefaultKwargs])}")

check_init = DebugStep(
    description="Checking it is possible to compute an initial state",
    attrs=(InitialState,),
    check_fn=_always_succeed,
    success_msg=lambda state: pytree_to_msg(state[InitialState],
                                            "Initial state", state[Verbosity]))


check_init_nan = DebugStep(
    description="Checking initial state is has no NaN",
    attrs=(InitialState,),
    check_fn=lambda model, debug_state, kwargs: ~_has_nans(
        debug_state[InitialState]),
    success_msg=lambda state: "No nans detected!",
    failure_msgs=[
        lambda state: pytree_to_msg(
            state[InitialState], "Initial state has nans!",
            state[Verbosity])])


def _check_init_log_density(model, debug_state, kwargs):
  del model, kwargs
  expected_shape = (_get_num_chains(debug_state[DefaultKwargs]),)
  right_shape = jnp.shape(debug_state[InitialStateLogDensity]) == expected_shape
  has_nans = ~_has_nans(debug_state[InitialStateLogDensity])
  return right_shape and has_nans


check_init_log_density = DebugStep(
    description="Computing initial state log density",
    attrs=(InitialStateLogDensity, DefaultKwargs),
    check_fn=_check_init_log_density,
    success_msg=lambda state: pytree_to_msg(state[
        InitialStateLogDensity], "Initial state log density", state[Verbosity]),
    failure_msgs=[
        lambda state:
        f"Initial state has log density\n{state[InitialStateLogDensity]!r}"
    ])


def _get_unconstrained_state(model, debug_state, kwargs):
  del kwargs
  debug_state["unconstrained_state"] = model.inverse_transform_fn(
      debug_state["initial_state"])
  return debug_state


def _check_unconstrained_state(model, debug_state, kwargs):
  del model, kwargs
  if _has_nans(debug_state[UnconstrainedState]):
    debug_state["_msg"] = (
        "Some points are NaN. Check `model.inverse_transform_fn`, which is "
        "automatically generated by `oryx.core.inverse(model.transform_fn)`, "
        "and should map from the support of the log density to R^n. Maybe "
        "the opposite direction transform was provided?")
    return False
  return True


check_transform = DebugStep(
    description="Transforming model to R^n",
    attrs=(UnconstrainedState,),
    check_fn=_check_unconstrained_state,
    exceptions=(NotImplementedError,),
    success_msg=lambda state: pytree_to_msg(state[
        UnconstrainedState], "Transformed state", state[Verbosity]),
    failure_msgs=[
        lambda state:
        "Automatic inverse and ildj not supported for this model.",
        lambda state: state["_msg"]
    ])


def _check_transformed_log_density_no_ildj(model, debug_state, kwargs):
  del model, kwargs
  expected_shape = (_get_num_chains(debug_state[DefaultKwargs]),)
  return jnp.shape(debug_state[UnconstrainedLogDensityNoILDJ]) == expected_shape


def _check_transformed_log_density(model, debug_state, kwargs):
  del model, kwargs
  expected_shape = (_get_num_chains(debug_state[DefaultKwargs]),)
  return jnp.shape(debug_state[UnconstrainedLogDensity]) == expected_shape


check_transformed_log_density_no_ildj = DebugStep(
    description="Computing transformed state log density shape",
    attrs=(UnconstrainedLogDensityNoILDJ, DefaultKwargs),
    check_fn=_check_transformed_log_density_no_ildj,
    success_msg=lambda state: pytree_to_msg(state[
        UnconstrainedLogDensityNoILDJ], "Transformed state log density", state[
            Verbosity]),
    failure_msgs=[
        lambda state: pytree_to_msg(state[UnconstrainedLogDensityNoILDJ],
                                    "Transformed state log density", state[
                                        Verbosity])
    ])

check_transformed_log_density = DebugStep(
    description="Computing transformed state log density shape",
    attrs=(UnconstrainedLogDensity, DefaultKwargs),
    check_fn=_check_transformed_log_density,
    success_msg=lambda state: pytree_to_msg(state[
        UnconstrainedLogDensity], "Transformed state log density", state[
            Verbosity]),
    failure_msgs=[
        lambda state: pytree_to_msg(state[UnconstrainedLogDensity],
                                    "Transformed state log density", state[
                                        Verbosity])
    ])


compute_gradients_no_ildj = DebugStep(
    description="Computing gradients of transformed log density",
    attrs=(InitialGradientsNoILDJ,),
    check_fn=lambda model, debug_state, _: ~_has_nans(
        debug_state[InitialGradientsNoILDJ]),
    success_msg=lambda state: pytree_to_msg(
        state[InitialGradientsNoILDJ], "Initial gradient", state[Verbosity]
    ),
    failure_msgs=[
        lambda state: "The gradient contains NaNs! "
        + pytree_to_msg(
            state[InitialGradientsNoILDJ], "Initial gradients", state[Verbosity]
        )
    ],
)
compute_gradients = DebugStep(
    description="Computing gradients of transformed log density",
    attrs=(InitialGradients,),
    check_fn=lambda model, debug_state, _: ~_has_nans(
        debug_state[InitialGradients]),
    success_msg=lambda state: pytree_to_msg(
        state[InitialGradients], "Initial gradient", state[Verbosity]
    ),
    failure_msgs=[
        lambda state: "The gradient contains NaNs! "
        + pytree_to_msg(
            state[InitialGradients], "Initial gradients", state[Verbosity]
        )
    ],
)
# pylint: enable=g-long-lambda


class ModelDebug:
  """Controller for running debug steps."""

  def __init__(self, model, seed, debug_steps, kwargs):
    self.model = model
    self.seed = seed
    self.debug_steps = debug_steps
    self.kwargs = kwargs

  def __call__(
      self, verbosity, catch_exceptions: bool, printer: Callable[..., Any]
  ):
    if verbosity <= 0:
      printer = lambda *args, **kwargs: None
    debug_state = {Seed: self.seed, Verbosity: verbosity}
    debug_print = make_debug_print(printer)
    all_passed = True
    for debug_step in self.debug_steps:
      passed, debug_state = debug_step(
          model=self.model,
          debug_state=debug_state,
          verbosity=verbosity,
          debug_print=debug_print,
          kwargs=self.kwargs,
          catch_exceptions=catch_exceptions)
      all_passed &= passed
      if verbosity < 3:
        msg = ""
      else:
        msg = "".join([_BAD, _GOOD][passed]() for _ in range(10)) + "\n"
      debug_print(msg=msg, verbosity=verbosity)
    return all_passed


def debug(model, seed, verbosity, printer, kwargs, catch_exceptions: bool):
  """Debugger that includes the inverse log det jacobian."""
  checkers = [
      check_shapes,
      check_test_point_log_density,
      check_kwargs,
      check_init,
      check_init_nan,
      check_init_log_density,
      check_transform,
      check_transformed_log_density,
      compute_gradients]
  if kwargs is None:
    kwargs = {}
  return ModelDebug(model, seed, checkers, kwargs)(
      verbosity=verbosity, catch_exceptions=catch_exceptions, printer=printer)


def debug_no_ildj(
    model, seed, verbosity, printer, kwargs, catch_exceptions: bool
):
  """Debugger for models with no inverse log det jacobian correction."""
  checkers = [
      check_shapes,
      check_test_point_log_density,
      check_kwargs,
      check_init,
      check_init_nan,
      check_init_log_density,
      check_transform,
      check_transformed_log_density_no_ildj,
      compute_gradients_no_ildj]
  if kwargs is None:
    kwargs = {}
  return ModelDebug(model, seed, checkers, kwargs)(
      verbosity=verbosity, catch_exceptions=catch_exceptions, printer=printer)


def _get_num_chains(default_kwargs):
  for v in default_kwargs.values():
    for key in ("num_chains", "num_particles", "batch_size"):
      if key in v:
        return v[key]
  raise KeyError("No `num_chains` in default kwargs!")


def _has_nans(pytree):
  return jax.tree_util.tree_reduce(lambda t, c: t and jnp.any(jnp.isnan(c)),
                                   pytree, True)



# Copyright 2024 The bayeux Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""BlackJAX specific code."""
import functools

import arviz as az
from bayeux._src import shared
import blackjax
import jax
import jax.numpy as jnp
import optax


_ADAPT_FNS = {
    "window": blackjax.window_adaptation,
    "pathfinder": blackjax.pathfinder_adaptation,
    "chees": blackjax.chees_adaptation,
    "meads": blackjax.meads_adaptation,
}

_ALGORITHMS = {
    "hmc": blackjax.hmc,
    "ghmc": blackjax.ghmc,
    "dynamic_hmc": blackjax.dynamic_hmc,
    "nuts": blackjax.nuts,
}


def _convert_algorithm(algorithm):
  # Remove this after blackjax is stable
  if hasattr(algorithm, "differentiable"):
    return algorithm.differentiable
  return algorithm


def get_extra_kwargs(kwargs):
  defaults = {
      "chain_method": "vectorized",
      "num_chains": 8,
      "num_draws": 500,
      "num_adapt_draws": 500,
      "return_pytree": False}
  shared.update_with_kwargs(defaults, kwargs=kwargs)
  return defaults


class _BlackjaxSampler(shared.Base):
  """Base class for blackjax samplers."""
  name: str = ""
  adapt_fn: str = ""
  algorithm: str = ""

  def get_kwargs(self, **kwargs):
    adapt_fn = _ADAPT_FNS[self.adapt_fn]
    algorithm = _ALGORITHMS[self.algorithm]
    extra_parameters = get_extra_kwargs(kwargs)
    constrained_log_density = self.constrained_log_density()
    adaptation_kwargs, run_kwargs = get_adaptation_kwargs(
        adapt_fn, algorithm, constrained_log_density, extra_parameters | kwargs)
    return {adapt_fn: adaptation_kwargs,
            "adapt.run": run_kwargs,
            _convert_algorithm(algorithm): get_algorithm_kwargs(
                _convert_algorithm(algorithm), constrained_log_density, kwargs),
            "extra_parameters": extra_parameters}

  def __call__(self, seed, **kwargs):
    init_key, sample_key = jax.random.split(seed)
    kwargs = self.get_kwargs(**kwargs)
    initial_state = self.get_initial_state(
        init_key, num_chains=kwargs["extra_parameters"]["num_chains"])

    return _sample_blackjax(
        initial_state=self.inverse_transform_fn(initial_state),
        algorithm=_ALGORITHMS[self.algorithm],
        transform_fn=self.transform_fn,
        adapt_fn=_ADAPT_FNS[self.adapt_fn],
        seed=sample_key,
        kwargs=kwargs)


class _BlackjaxDynamicSampler(_BlackjaxSampler):
  """Base class for blackjax samplers."""

  def __call__(self, seed, **kwargs):
    init_key, sample_key = jax.random.split(seed)
    kwargs = self.get_kwargs(**kwargs)
    initial_state = self.get_initial_state(
        init_key, num_chains=kwargs["extra_parameters"]["num_chains"])

    return _sample_blackjax_dynamic(
        initial_state=self.inverse_transform_fn(initial_state),
        algorithm=_ALGORITHMS[self.algorithm],
        transform_fn=self.transform_fn,
        adapt_fn=_ADAPT_FNS[self.adapt_fn],
        seed=sample_key,
        kwargs=kwargs)


class HMC(_BlackjaxSampler):
  name = "blackjax_hmc"
  adapt_fn = "window"
  algorithm = "hmc"


class CheesHMC(_BlackjaxDynamicSampler):
  name = "blackjax_chees_hmc"
  adapt_fn = "chees"
  algorithm = "dynamic_hmc"


class MeadsHMC(_BlackjaxDynamicSampler):
  name = "blackjax_meads_hmc"
  adapt_fn = "meads"
  algorithm = "ghmc"


class HMCPathfinder(_BlackjaxSampler):
  name = "blackjax_hmc_pathfinder"
  adapt_fn = "pathfinder"
  algorithm = "hmc"


class NUTS(_BlackjaxSampler):
  name = "blackjax_nuts"
  adapt_fn = "window"
  algorithm = "nuts"


class NUTSPathfinder(_BlackjaxSampler):
  name = "blackjax_nuts_pathfinder"
  adapt_fn = "pathfinder"
  algorithm = "nuts"


def _blackjax_adapt(
    seed,
    adapt_fn,
    kwarg_dict,
    **kwargs):
  adapt = adapt_fn(**kwarg_dict[adapt_fn])
  (last_state, parameters), _ = adapt.run(
      rng_key=seed, **kwargs,
      **kwarg_dict["adapt.run"])
  return last_state, parameters


def _blackjax_inference(
    seed,
    adapt_state,
    adapt_parameters,
    algorithm,
    num_draws,
    kwargs):
  """Run blackjax inference loop in a vmappable way.

  Args:
    seed: jax PRNGKey
    adapt_state: return value from a blackjax adaptation algorithm.
    adapt_parameters: return value from blackjax adaptation algorithm
    algorithm: name of algorithm to run.
    num_draws: number of iterations to run for.
    kwargs: Extra keyword arguments for algorithm.

  Returns:
  The results of the inference and the adaptation:
    (states, infos), adaptation_parameters
  """

  algorithm_kwargs = kwargs[_convert_algorithm(algorithm)] | adapt_parameters
  inference_algorithm = algorithm(**algorithm_kwargs)
  _, states, infos = blackjax.util.run_inference_algorithm(
      rng_key=seed,
      initial_state_or_position=adapt_state,
      inference_algorithm=inference_algorithm,
      num_steps=num_draws,
      progress_bar=False)
  return states, infos


def _blackjax_inference_loop(
    seed,
    init_position,
    adapt_fn,
    algorithm,
    num_draws,
    kwargs):
  """Constructs and runs inference loop."""
  adapt_seed, inference_seed = jax.random.split(seed)
  adapt_state, adapt_parameters = _blackjax_adapt(
      adapt_seed, adapt_fn, kwarg_dict=kwargs, position=init_position)
  return _blackjax_inference(
      inference_seed,
      adapt_state,
      adapt_parameters,
      algorithm,
      num_draws,
      kwargs), adapt_parameters


def _blackjax_stats_to_dict(sample_stats, potential_energy, adapt_parameters):
  """Extract ArviZ compatible stats from blackjax sampler.

  Adapted from https://github.com/pymc-devs/pymc

  Args:
    sample_stats: Blackjax NUTSInfo object containing sampler statistics.
    potential_energy: Potential energy values of sampled positions.
    adapt_parameters: Parameters from adaptation.

  Returns:
      Dictionary of sampler statistics.
  """
  rename_key = {
      "is_divergent": "diverging",
      "energy": "energy",
      "num_trajectory_expansions": "tree_depth",
      "num_integration_steps": "n_steps",
      "acceptance_rate": "acceptance_rate",         # naming here depends
      "acceptance_probability": "acceptance_rate",  # on blackjax version
  }
  converted_stats = {"lp": potential_energy}
  step_size = adapt_parameters.get("step_size", None)
  if step_size is not None:
    if jnp.ndim(step_size) == 0:
      converted_stats["step_size"] = jnp.full_like(potential_energy, step_size)
    else:
      converted_stats["step_size"] = jnp.repeat(
          step_size[..., None], repeats=jnp.shape(potential_energy)[-1], axis=-1
      )
  for old_name, new_name in rename_key.items():
    value = getattr(sample_stats, old_name, None)
    if value is not None:
      converted_stats[new_name] = value
  return converted_stats


def get_adaptation_kwargs(adaptation_algorithm, algorithm, log_density, kwargs):
  """Sets defaults and merges user-provided adaptation keywords."""
  adaptation_kwargs, adaptation_required = shared.get_default_signature(
      adaptation_algorithm)
  adaptation_kwargs.update(
      {k: kwargs[k] for k in adaptation_required if k in kwargs})
  if "logdensity_fn" in adaptation_required:
    adaptation_kwargs["logdensity_fn"] = log_density
    adaptation_required.remove("logdensity_fn")
  elif "logprob_fn" in adaptation_required:
    adaptation_kwargs["logprob_fn"] = log_density
    adaptation_required.remove("logprob_fn")

  adaptation_required.discard("extra_parameters")
  if "algorithm" in adaptation_required:
    adaptation_required.remove("algorithm")
    adaptation_kwargs["algorithm"] = algorithm
    adaptation_kwargs = (
        get_algorithm_kwargs(_convert_algorithm(algorithm), log_density, kwargs)
        | adaptation_kwargs)

  adaptation_required = adaptation_required - adaptation_kwargs.keys()

  if adaptation_required:
    raise ValueError(
        "Unexpected required arguments: "
        f"{','.join(adaptation_required)}. Probably file a bug, but "
        "you can try to manually supply them as keywords."
    )
  adaptation_kwargs.update(
      {k: kwargs[k] for k in adaptation_kwargs if k in kwargs}
  )
  # step_size will get adapted -- maybe warn if this is set manually, and
  # suggest setting init_step_size instead?
  adaptation_kwargs.pop("step_size", None)
  # blackjax doesn't have a pleasant way to accept this argument --
  # window_adaptation calls `algorithm.build_kernel()` with no arguments, but
  # it should probably take the below arguments:
  adaptation_kwargs.pop("divergence_threshold", None)
  adaptation_kwargs.pop("integrator", None)
  adaptation_kwargs.pop("max_num_doublings", None)

  adapt = adaptation_algorithm(**adaptation_kwargs)
  run_kwargs, run_required = shared.get_default_signature(adapt.run)
  run_required.remove("rng_key")
  run_kwargs.update({k: kwargs[k] for k in run_required if k in kwargs})
  if "optim" in run_required:
    run_kwargs["optim"] = optax.adam(learning_rate=0.01)
    run_required.remove("optim")
  if "step_size" in run_required:
    run_kwargs["step_size"] = 0.5
    run_required.remove("step_size")
  run_kwargs["num_steps"] = kwargs.get("num_adapt_draws",
                                       run_kwargs["num_steps"])

  return adaptation_kwargs, run_kwargs


def get_algorithm_kwargs(algorithm, log_density, kwargs):
  """Sets defaults and merges user-provided keywords for sampling."""
  algorithm_kwargs, algorithm_required = shared.get_default_signature(algorithm)
  kwargs_with_defaults = {
      "logdensity_fn": log_density,
      "step_size": 0.5,
      "num_integration_steps": 16,
  } | kwargs
  shared.update_with_kwargs(
      algorithm_kwargs, reqd=algorithm_required, kwargs=kwargs_with_defaults)
  algorithm_required.remove("logdensity_fn")
  algorithm_required.discard("inverse_mass_matrix")
  algorithm_required.discard("alpha")
  algorithm_required.discard("delta")
  algorithm_required.discard("momentum_inverse_scale")

  algorithm_required = algorithm_required - algorithm_kwargs.keys()
  if algorithm_required:
    raise ValueError(f"Unexpected required arguments: "
                     f"{','.join(algorithm_required)}. Probably file a bug, but"
                     " you can try to manually supply them as keywords.")
  return algorithm_kwargs


def _sample_blackjax_dynamic(
    *,
    initial_state,
    algorithm,
    seed,
    transform_fn,
    adapt_fn,
    kwargs):
  """Constructs and runs blackjax sampler."""
  extra_parameters = kwargs.pop("extra_parameters")
  num_draws = extra_parameters["num_draws"]
  num_chains = extra_parameters["num_chains"]
  chain_method = extra_parameters["chain_method"]

  adapt_seed, seed = jax.random.split(seed)
  adapt_state, adapt_parameters = _blackjax_adapt(
      seed=adapt_seed,
      adapt_fn=adapt_fn,
      kwarg_dict=kwargs,
      positions=initial_state,
  )
  sampler = functools.partial(
      _blackjax_inference,
      adapt_parameters=adapt_parameters,
      algorithm=algorithm,
      num_draws=num_draws,
      kwargs=kwargs)
  map_seed = jax.random.split(seed, num_chains)
  mapped_sampler = shared.map_fn(chain_method, sampler)

  states, stats = mapped_sampler(map_seed, adapt_state)
  draws = transform_fn(states.position)
  if extra_parameters["return_pytree"]:
    return draws
  else:
    potential_energy = states.logdensity
    sample_stats = _blackjax_stats_to_dict(
        stats, potential_energy, adapt_parameters)
    if hasattr(draws, "_asdict"):
      draws = draws._asdict()
    elif not isinstance(draws, dict):
      draws = {"var0": draws}
    return az.from_dict(posterior=draws, sample_stats=sample_stats)


def _sample_blackjax(
    *,
    initial_state,
    algorithm,
    seed,
    transform_fn,
    adapt_fn,
    kwargs):
  """Constructs and runs blackjax sampler."""
  extra_parameters = kwargs.pop("extra_parameters")
  num_draws = extra_parameters["num_draws"]
  num_chains = extra_parameters["num_chains"]
  chain_method = extra_parameters["chain_method"]
  sampler = functools.partial(
      _blackjax_inference_loop,
      algorithm=algorithm,
      adapt_fn=adapt_fn,
      num_draws=num_draws,
      kwargs=kwargs)
  map_seed = jax.random.split(seed, num_chains)
  mapped_sampler = shared.map_fn(chain_method, sampler)

  (states, stats), adapt_parameters = mapped_sampler(map_seed, initial_state)
  draws = transform_fn(states.position)
  if extra_parameters["return_pytree"]:
    return draws
  else:
    potential_energy = states.logdensity
    sample_stats = _blackjax_stats_to_dict(
        stats, potential_energy, adapt_parameters)
    if hasattr(draws, "_asdict"):
      draws = draws._asdict()
    elif not isinstance(draws, dict):
      draws = {"var0": draws}
    return az.from_dict(posterior=draws, sample_stats=sample_stats)



# Copyright 2024 The bayeux Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Shared functions for optimizers."""
import collections

from bayeux._src import debug
from bayeux._src import shared


OptimizerResults = collections.namedtuple("OptimizerResults",
                                          ["params", "state", "loss"])


def get_extra_kwargs(kwargs):
  return {
      "chain_method": kwargs.get("chain_method", "vectorized"),
      "num_particles": kwargs.get("num_particles", 8),
      "num_iters": kwargs.get("num_iters", 1000),
      "apply_transform": kwargs.get("apply_transform", True),
  }


def get_optimizer_kwargs(optimizer, kwargs, ignore_required=None):
  """Sets defaults and merges user-provided adaptation keywords."""
  if ignore_required is None:
    ignore_required = set()
  optimizer_kwargs, optimizer_required = shared.get_default_signature(
      optimizer)
  optimizer_kwargs.update(
      {k: kwargs[k] for k in optimizer_required if k in kwargs})

  optimizer_required = (optimizer_required -
                        ignore_required -
                        set(optimizer_kwargs.keys()))

  if optimizer_required:
    raise ValueError(
        "Unexpected required arguments: "
        f"{','.join(optimizer_required)}. Probably file a bug, but "
        "you can try to manually supply them as keywords."
    )
  optimizer_kwargs.update(
      {k: kwargs[k] for k in optimizer_kwargs if k in kwargs})
  return optimizer_kwargs


class Optimizer(shared.Base):
  """Base class for optimizers."""
  name: str = ""
  optimizer: str = ""

  def debug(
      self,
      seed,
      verbosity=2,
      catch_exceptions: bool = True,
      printer=print,
      kwargs=None,
  ):
    return debug.debug_no_ildj(
        self,
        seed,
        verbosity=verbosity,
        catch_exceptions=catch_exceptions,
        printer=printer,
        kwargs=kwargs,
    )

  def default_kwargs(self):
    return {"learning_rate": 0.1}

  def negative_log_prob(self):
    return lambda x: -self.log_density(x)

  def transformed_negative_log_prob(self):
    return lambda x: -self.log_density(self.transform_fn(x))

  def _map_optimizer(self, chain_method, fit):
    return shared.map_fn(chain_method, fit)

  def _prep_args(self, seed, kwargs):
    num_particles = kwargs["extra_parameters"]["num_particles"]
    initial_state = self.get_initial_state(seed, num_chains=num_particles)
    apply_transform = kwargs["extra_parameters"]["apply_transform"]
    if apply_transform:
      fun = self.transformed_negative_log_prob()
      initial_state = self.inverse_transform_fn(initial_state)
    else:
      fun = self.negative_log_prob()
    return fun, initial_state, apply_transform



# Copyright 2024 The bayeux Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TensorF[riendly] Probability specific code."""
import functools

from bayeux._src import shared
import jax
import jax.numpy as jnp
import optax
import oryx
import tensorflow_probability.substrates.jax as tfp

tfb = tfp.bijectors


class Custom(tfb.Bijector):
  """Custom tfp bijector."""

  def __init__(self, bx_model):
    super().__init__(
        forward_min_event_ndims=jax.tree.map(jnp.ndim, bx_model.test_point))
    self.bx_model = bx_model

  def _forward(self, x):
    return self.bx_model.transform_fn(x)

  def _inverse(self, y):
    return self.bx_model.inverse_transform_fn(y)

  def _inverse_log_det_jacobian(self, y):
    return self.bx_model.inverse_log_det_jacobian(y)

  def _forward_log_det_jacobian(self, x):
    return -self.inverse_log_det_jacobian(self.forward(x))

  def _forward_event_shape_tensor(self, input_shape):
    return jax.tree.map(jnp.shape,
                        self._forward(jax.tree.map(jnp.ones, input_shape)))

  def _inverse_event_shape_tensor(self, output_shape):
    return jax.tree.map(jnp.shape,
                        self._inverse(jax.tree.map(jnp.ones, output_shape)))


def get_fit_kwargs(log_density, kwargs):
  """Get keyword arguments for fitting VI."""
  fit_kwargs, fit_required = shared.get_default_signature(
      tfp.vi.fit_surrogate_posterior_stateless)
  fit_kwargs.pop("seed")
  fit_kwargs["optimizer"] = optax.adam(learning_rate=0.01)
  fit_kwargs["target_log_prob_fn"] = jax.vmap(log_density)

  fit_kwargs = {
      "sample_size": 16,
      "num_steps": 2_000,
      "jit_compile": True,
  } | fit_kwargs
  fit_kwargs.update({k: kwargs[k] for k in fit_required if k in kwargs})
  fit_required = (fit_required
                  - set(fit_kwargs.keys())
                  - {"build_surrogate_posterior_fn", "initial_parameters"})
  if fit_required:
    raise ValueError(
        "Unexpected required arguments: "
        f"{','.join(fit_required)}. Probably file a bug, but "
        "you can try to manually supply them as keywords.")
  return fit_kwargs


def get_build_kwargs(event_shape, bijector, kwargs):
  """Get keyword arguments for building VI."""
  build_kwargs, build_required = shared.get_default_signature(
      tfp.experimental.vi.build_factored_surrogate_posterior_stateless)
  build_kwargs["event_shape"] = event_shape
  build_kwargs["bijector"] = bijector
  build_kwargs.update({k: kwargs[k] for k in build_required if k in kwargs})
  build_required = build_required - set(build_kwargs.keys())
  if build_required:
    raise ValueError(
        "Unexpected required arguments: "
        f"{','.join(build_required)}. Probably file a bug, but "
        "you can try to manually supply them as keywords."
    )
  return build_kwargs


class Factored(shared.Base):
  """Base class for TFP variational inference."""
  name = "tfp_factored_surrogate_posterior"

  def get_kwargs(self, **kwargs):
    return {
        tfp.experimental.vi.build_factored_surrogate_posterior_stateless: (
            get_build_kwargs(
                jax.tree.map(jnp.shape, self.test_point),
                self.constraining_bijector(),
                kwargs)),
        tfp.vi.fit_surrogate_posterior_stateless: get_fit_kwargs(
            self.log_density, kwargs),
        "extra_parameters": {
            "chain_method": kwargs.get("chain_method", "vectorized"),
            "batch_size": kwargs.get("batch_size", 16),
        }}

  def constraining_bijector(self):
    if self.inverse_log_det_jacobian is None:
      self.inverse_log_det_jacobian = oryx.core.ildj(self.transform_fn)
    return Custom(self)

  def __call__(self, seed, **kwargs):
    kwargs = self.get_kwargs(**kwargs)
    batch_size = kwargs["extra_parameters"]["batch_size"]
    # TODO(colcarroll): The initialization is being done by TFP here. It would
    # be nice to put a vmap-friendly initialization in.
    init_fn, build_surrogate_posterior_fn = (
        tfp.experimental.vi.build_factored_surrogate_posterior_stateless(
            **kwargs[
                tfp.experimental.vi.build_factored_surrogate_posterior_stateless
            ]))

    fit_fn = _make_tfp_fit(
        build_surrogate_posterior_fn=build_surrogate_posterior_fn,
        init_fn=init_fn,
        fit_kwargs=kwargs[tfp.vi.fit_surrogate_posterior_stateless])
    chain_method = kwargs["extra_parameters"]["chain_method"]
    if chain_method == "vectorized":
      mapped_fit = jax.vmap(fit_fn)
    elif chain_method == "parallel":
      mapped_fit = jax.pmap(fit_fn)
    elif chain_method == "sequential":
      mapped_fit = functools.partial(jax.tree.map, fit_fn)
    else:
      raise ValueError(f"Chain method {chain_method} not supported.")

    map_seed = jax.random.split(seed, batch_size)
    params, loss = mapped_fit(seed=map_seed)
    return build_surrogate_posterior_fn(params), loss


def _make_tfp_fit(
    build_surrogate_posterior_fn, init_fn, fit_kwargs
):
  def run_vi(seed):
    init_seed, opt_seed = jax.random.split(seed)
    return tfp.vi.fit_surrogate_posterior_stateless(
        build_surrogate_posterior_fn=build_surrogate_posterior_fn,
        initial_parameters=init_fn(init_seed),
        seed=opt_seed,
        **fit_kwargs)

  return run_vi



# Copyright 2024 The bayeux Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2024 The bayeux Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TFP specific code."""

import arviz as az
from bayeux._src import shared
import jax
import jax.numpy as jnp
import numpy as np
import tensorflow_probability.substrates.jax as tfp


_ALGORITHMS = {
    "hmc": tfp.mcmc.HamiltonianMonteCarlo,
    "nuts": tfp.mcmc.NoUTurnSampler,
}

_TRACE_FNS = {
    "hmc": tfp.experimental.mcmc.windowed_sampling.default_hmc_trace_fn,
    "nuts": tfp.experimental.mcmc.windowed_sampling.default_nuts_trace_fn,
}


class SnaperHMC(shared.Base):
  """Implements SNAPER HMC [1] with step size adaptation.

  [1]: Sountsov, P. & Hoffman, M. (2021). Focusing on Difficult Directions for
  Learning HMC Trajectory Lengths. <https://arxiv.org/abs/2110.11576>
  """
  name = "tfp_snaper_hmc"

  def get_kwargs(self, **kwargs):
    kwargs_with_defaults = {
        "num_results": 1_000,
        "num_chains": 8,
    } | kwargs
    snaper = tfp.experimental.mcmc.sample_snaper_hmc
    snaper_kwargs, snaper_required = shared.get_default_signature(snaper)
    snaper_kwargs.update({k: kwargs_with_defaults[k] for k in snaper_required
                          if k in kwargs_with_defaults})
    snaper_required.remove("model")
    # Initial state is handled internally
    snaper_kwargs.pop("init_state")
    # Seed set later
    snaper_kwargs.pop("seed")

    snaper_required = snaper_required - snaper_kwargs.keys()

    if snaper_required:
      raise ValueError(f"Unexpected required arguments: "
                       f"{','.join(snaper_required)}. Probably file a bug, but "
                       "you can try to manually supply them as keywords.")
    snaper_kwargs.update({k: kwargs_with_defaults[k] for k in snaper_kwargs
                          if k in kwargs_with_defaults})
    return {
        snaper: snaper_kwargs,
        "extra_parameters": {
            "return_pytree": kwargs.get("return_pytree", False)
        },
    }

  def __call__(self, seed, **kwargs):
    snaper = tfp.experimental.mcmc.sample_snaper_hmc
    init_key, sample_key = jax.random.split(seed)
    kwargs = self.get_kwargs(**kwargs)
    initial_state = self.get_initial_state(
        init_key, num_chains=kwargs[snaper]["num_chains"])

    vmapped_constrained_log_prob = jax.vmap(self.constrained_log_density())

    def tlp(*args, **kwargs):
      if args:
        return vmapped_constrained_log_prob(args)
      else:
        return vmapped_constrained_log_prob(kwargs)

    (draws, trace), *_ = snaper(
        model=tlp, init_state=initial_state, seed=sample_key, **kwargs[snaper]
    )
    draws = self.transform_fn(draws)
    if kwargs["extra_parameters"]["return_pytree"]:
      return draws

    if hasattr(draws, "_asdict"):
      draws = draws._asdict()
    elif not isinstance(draws, dict):
      draws = {"var0": draws}

    draws = {x: np.swapaxes(v, 0, 1) for x, v in draws.items()}
    return az.from_dict(posterior=draws, sample_stats=_tfp_stats_to_dict(trace))


class _TFPBase(shared.Base):
  """Base class for TFP windowed samplers."""
  name: str = ""
  algorithm: str = ""

  def get_kwargs(self, **kwargs):
    if self.algorithm == "nuts":
      target_accept_prob = 0.8
    else:
      target_accept_prob = 0.6

    kwargs = {
        "target_accept_prob": target_accept_prob,
        "reduce_fn": tfp.math.reduce_log_harmonic_mean_exp,
        "num_adaptation_steps": 500,
        "step_size": 0.5,
        "num_leapfrog_steps": 8,
    } | kwargs
    extra_parameters = {
        "num_draws": 1_000,
        "num_chains": 8,
        "num_adaptation_steps": 500,
        "return_pytree": False,
    }
    shared.update_with_kwargs(extra_parameters, kwargs=kwargs)

    dual_averaging_kwargs, da_reqd = shared.get_default_signature(
        tfp.mcmc.DualAveragingStepSizeAdaptation)
    shared.update_with_kwargs(
        dual_averaging_kwargs, reqd=da_reqd, kwargs=kwargs)
    da_reqd = da_reqd - dual_averaging_kwargs.keys()
    da_reqd.remove("inner_kernel")
    if da_reqd:
      raise ValueError(
          "Unexpected required arguments: "
          f"{','.join(da_reqd)}. Probably file a bug, but "
          "you can try to manually supply them as keywords."
      )

    proposal_kwargs, proposal_reqd = shared.get_default_signature(
        _ALGORITHMS[self.algorithm])
    shared.update_with_kwargs(
        proposal_kwargs, reqd=proposal_reqd, kwargs=kwargs)
    proposal_reqd = proposal_reqd - proposal_kwargs.keys()
    proposal_reqd.remove("target_log_prob_fn")
    if proposal_reqd:
      raise ValueError(
          "Unexpected required arguments: "
          f"{','.join(proposal_reqd)}. Probably file a bug, but "
          "you can try to manually supply them as keywords."
      )
    return {
        "extra_parameters": extra_parameters,
        "dual_averaging_kwargs": dual_averaging_kwargs,
        "proposal_kernel_kwargs": proposal_kwargs}

  def __call__(self, seed, **kwargs):
    kwargs = self.get_kwargs(**kwargs)
    init_key, sample_key = jax.random.split(seed)

    extra_parameters = kwargs["extra_parameters"]
    dual_averaging_kwargs = kwargs["dual_averaging_kwargs"]
    proposal_kernel_kwargs = kwargs["proposal_kernel_kwargs"]
    initial_state = self.get_initial_state(
        init_key, num_chains=extra_parameters["num_chains"])

    vmapped_constrained_log_density = jax.vmap(self.constrained_log_density())
    initial_transformed_position, treedef = jax.tree_util.tree_flatten(
        self.inverse_transform_fn(initial_state))

    def target_log_prob_fn(*args):
      return vmapped_constrained_log_density(
          jax.tree_util.tree_unflatten(treedef, args))
    proposal_kernel_kwargs["target_log_prob_fn"] = target_log_prob_fn

    initial_running_variance = [
        tfp.experimental.stats.sample_stats.RunningVariance.from_stats(
            num_samples=jnp.array(1, part.dtype),
            mean=jnp.zeros_like(part),
            variance=jnp.ones_like(part))
        for part in initial_transformed_position]

    # The public API expects a JointDistribution. Much of the above is adapted
    # from the source code for
    # `tfp.experimental.mcmc.windowed_adaptive_{nuts|hmc}`, but handling a raw
    # log density, and doing the structure flattening with `jax.tree_utils`.
    draws, trace = tfp.experimental.mcmc.windowed_sampling._do_sampling(
        kind=self.algorithm,
        proposal_kernel_kwargs=proposal_kernel_kwargs,
        dual_averaging_kwargs=dual_averaging_kwargs,
        num_draws=extra_parameters["num_draws"],
        num_burnin_steps=extra_parameters["num_adaptation_steps"],
        initial_position=initial_transformed_position,
        initial_running_variance=initial_running_variance,
        bijector=None,
        trace_fn=_TRACE_FNS[self.algorithm],
        return_final_kernel_results=False,
        chain_axis_names=None,
        shard_axis_names=None,
        seed=sample_key)

    draws = self.transform_fn(jax.tree_util.tree_unflatten(treedef, draws))
    if extra_parameters["return_pytree"]:
      return draws

    if hasattr(draws, "_asdict"):
      draws = draws._asdict()
    elif not isinstance(draws, dict):
      draws = {"var0": draws}

    draws = {x: np.swapaxes(v, 0, 1) for x, v in draws.items()}
    return az.from_dict(posterior=draws, sample_stats=_tfp_stats_to_dict(trace))


class NUTS(_TFPBase):
  name = "tfp_nuts"
  algorithm = "nuts"


class HMC(_TFPBase):
  name = "tfp_hmc"
  algorithm = "hmc"


def _tfp_stats_to_dict(stats):
  new_stats = {}
  for k, v in stats.items():
    if k == "variance_scaling":
      continue
    if np.ndim(v) > 1:
      new_stats[k] = np.swapaxes(v, 0, 1)
    else:
      new_stats[k] = v
  return new_stats



# Copyright 2024 The bayeux Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Imports from submodules."""
# pylint: disable=g-importing-member
# pylint: disable=g-import-not-at-top
import importlib.util

__all__ = []

if importlib.util.find_spec("jaxopt") is not None:
  from bayeux._src.optimize.jaxopt import BFGS
  from bayeux._src.optimize.jaxopt import GradientDescent
  from bayeux._src.optimize.jaxopt import LBFGS
  from bayeux._src.optimize.jaxopt import NonlinearCG
  __all__.extend(["BFGS", "GradientDescent", "LBFGS", "NonlinearCG"])

if importlib.util.find_spec("optax") is not None:
  from bayeux._src.optimize.optax import AdaBelief
  from bayeux._src.optimize.optax import Adafactor
  from bayeux._src.optimize.optax import Adagrad
  from bayeux._src.optimize.optax import Adam
  from bayeux._src.optimize.optax import Adamax
  from bayeux._src.optimize.optax import Adamw
  from bayeux._src.optimize.optax import Amsgrad
  # from bayeux._src.optimize.optax import Dpsgd  # pylint: disable=line-too-long
  from bayeux._src.optimize.optax import Fromage
  from bayeux._src.optimize.optax import Lamb
  # from bayeux._src.optimize.optax import Lars
  from bayeux._src.optimize.optax import Lion
  from bayeux._src.optimize.optax import NoisySgd
  from bayeux._src.optimize.optax import Novograd
  # from bayeux._src.optimize.optax import OptimisticGradientDescent  # pylint: disable=line-too-long
  from bayeux._src.optimize.optax import Radam
  from bayeux._src.optimize.optax import Rmsprop
  if importlib.util.find_spec("optax.contrib._schedule_free") is not None:
    from bayeux._src.optimize.optax import ScheduleFree
    __all__.append("ScheduleFree")
  from bayeux._src.optimize.optax import Sgd
  from bayeux._src.optimize.optax import Sm3
  from bayeux._src.optimize.optax import Yogi

  __all__.extend([
      "AdaBelief",
      "Adafactor",
      "Adagrad",
      "Adam",
      "Adamw",
      "Adamax",
      "Amsgrad",
      "Fromage",
      "Lamb",
      # "Lars",
      "Lion",
      "NoisySgd",
      "Novograd",
      # "OptimisticGradientDescent",
      # "Dpsgd",
      "Radam",
      "Rmsprop",
      "Sgd",
      "Sm3",
      "Yogi",
      ])

if importlib.util.find_spec("optimistix") is not None:
  from bayeux._src.optimize.optimistix import BFGS as optimistix_BFGS
  from bayeux._src.optimize.optimistix import NelderMead
  from bayeux._src.optimize.optimistix import NonlinearCG as optimistix_NonlinearCG

  __all__.extend([
      "optimistix_BFGS",
      "NelderMead",
      "optimistix_NonlinearCG"])



# Copyright 2024 The bayeux Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""NumPyro specific code."""

import arviz as az
from bayeux._src import shared
import jax
import numpyro.infer


def _potential(log_density):

  def mapped(*args, **kwargs):
    return -log_density(*args, **kwargs)

  return mapped

_ALGORITHMS = {"hmc": numpyro.infer.HMC, "nuts": numpyro.infer.NUTS}


class _NumpyroSampler(shared.Base):
  """Base class for NumPyro samplers."""
  name: str = ""
  algorithm: str = ""

  def get_kwargs(self, **kwargs):
    algorithm = _ALGORITHMS[self.algorithm]
    return {
        algorithm: get_sampler_kwargs(algorithm, kwargs),
        numpyro.infer.MCMC: get_mcmc_kwargs(kwargs),
        "extra_parameters": {
            "return_pytree": kwargs.get("return_pytree", False)
        },
    }

  def __call__(self, seed, **kwargs):
    init_key, sample_key = jax.random.split(seed)
    kwargs = self.get_kwargs(**kwargs)
    initial_state = self.get_initial_state(
        init_key, num_chains=kwargs[numpyro.infer.MCMC]["num_chains"])

    potential_fn = _potential(self.constrained_log_density())
    return _sample_numpyro(
        potential_fn=potential_fn,
        initial_state=initial_state,
        transform_fn=self.transform_fn,
        inverse_transform_fn=self.inverse_transform_fn,
        algorithm=_ALGORITHMS[self.algorithm],
        seed=sample_key,
        kwargs=kwargs)


class HMC(_NumpyroSampler):
  name = "numpyro_hmc"
  algorithm = "hmc"


class NUTS(_NumpyroSampler):
  name = "numpyro_nuts"
  algorithm = "nuts"


def get_sampler_kwargs(algorithm, kwargs):
  """Construct default args and include user arguments for samplers."""
  sampler_kwargs, sampler_required = shared.get_default_signature(algorithm)
  shared.update_with_kwargs(
      sampler_kwargs, reqd=sampler_required, kwargs=kwargs)
  sampler_kwargs.pop("potential_fn")

  sampler_required = sampler_required - sampler_kwargs.keys()

  if sampler_required:
    raise ValueError(f"Unexpected required arguments: "
                     f"{','.join(sampler_required)}. Probably file a bug, but "
                     "you can try to manually supply them as keywords.")
  return sampler_kwargs


def get_mcmc_kwargs(kwargs):
  """Construct default args and include user arguments for MCMC."""
  mcmc_kwargs, mcmc_required = shared.get_default_signature(numpyro.infer.MCMC)
  kwargs_with_defaults = {
      "num_warmup": 500,
      "num_samples": 1000,
      "num_chains": 8,
      "chain_method": "vectorized",
  } | kwargs
  shared.update_with_kwargs(
      mcmc_kwargs, reqd=mcmc_required, kwargs=kwargs_with_defaults)
  mcmc_required = mcmc_required - mcmc_kwargs.keys()
  mcmc_required.remove("sampler")
  if mcmc_required:
    raise ValueError(f"Unexpected required arguments: "
                     f"{','.join(mcmc_required)}. Probably file a bug, but "
                     "you can try to manually supply them as keywords.")
  return mcmc_kwargs


def _sample_numpyro(*, potential_fn, initial_state, transform_fn,
                    inverse_transform_fn, algorithm, seed, kwargs):
  """Run MCMC using NumPyro."""
  kernel = algorithm(
      potential_fn=potential_fn,
      **kwargs[algorithm],
  )

  mcmc_kernel = numpyro.infer.MCMC(
      sampler=kernel,
      **kwargs[numpyro.infer.MCMC],
  )

  init_params = inverse_transform_fn(initial_state)
  mcmc_kernel.run(
      seed,
      init_params=init_params,
      extra_fields=(
          "num_steps",
          "potential_energy",
          "energy",
          "adapt_state.step_size",
          "accept_prob",
          "diverging",
      ),
  )
  samples = mcmc_kernel.get_samples(group_by_chain=True)
  if kwargs["extra_parameters"]["return_pytree"]:
    return transform_fn(samples)
  else:
    mcmc_kernel.get_samples = lambda group_by_chain=True: transform_fn(samples)
    return az.from_numpyro(mcmc_kernel)



# Copyright 2024 The bayeux Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""bayeux API."""

# A new PyPI release will be pushed everytime `__version__` is increased
# When changing this, also update the CHANGELOG.md
__version__ = '0.1.12'

# Note: import <name> as <name> is required for names to be exported.
# See PEP 484 & https://github.com/google/jax/issues/7570
# pylint: disable=useless-import-alias
from bayeux import mcmc as mcmc
from bayeux import optimize as optimize
from bayeux._src import debug as debug
from bayeux._src import initialization as initialization
from bayeux._src import shared as shared
from bayeux._src import types as types
from bayeux._src.bayeux import Model  as Model  # pylint: disable=g-importing-member



# Copyright 2024 The bayeux Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2024 The bayeux Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""flowMC specific code."""
import arviz as az
from bayeux._src import shared
from flowMC import Sampler
from flowMC.nfmodel import realNVP
from flowMC.nfmodel import rqSpline
from flowMC.proposal import HMC
from flowMC.proposal import MALA
import jax
import jax.numpy as jnp


_NF_MODELS = {
    "real_nvp": realNVP.RealNVP,
    "masked_coupling_rq_spline": rqSpline.MaskedCouplingRQSpline,
}

_LOCAL_SAMPLERS = {"mala": MALA.MALA, "hmc": HMC.HMC}


def get_nf_model_kwargs(nf_model, n_features, kwargs):
  """Sets defaults and merges user-provided adaptation keywords."""
  defaults = {
      # RealNVP kwargs
      "n_hidden": 100,
      "n_layer": 10,
      # MaskedCouplingRQSpline kwargs
      "n_layers": 4,
      "num_bins": 8,
      "hidden_size": [64, 64],
      "spline_range": (-10.0, 10.0),
      "n_features": n_features,
  } | kwargs

  nf_model_kwargs, nf_model_required = shared.get_default_signature(
      nf_model)
  nf_model_kwargs.update(
      {k: defaults[k] for k in nf_model_required if k in defaults})
  nf_model_required.remove("key")
  nf_model_required.remove("kwargs")
  nf_model_required = nf_model_required - nf_model_kwargs.keys()

  if nf_model_required:
    raise ValueError(
        "Unexpected required arguments: "
        f"{','.join(nf_model_required)}. Probably file a bug, but "
        "you can try to manually supply them as keywords."
    )
  nf_model_kwargs.update(
      {k: defaults[k] for k in nf_model_kwargs if k in defaults})

  return nf_model_kwargs


def get_local_sampler_kwargs(local_sampler, log_density, n_features, kwargs):
  """Sets defaults and merges user-provided adaptation keywords."""

  defaults = {
      # HMC kwargs
      "condition_matrix": jnp.eye(n_features),
      "n_leapfrog": 10,
      # Both
      "step_size": 0.1,
      "logpdf": log_density
  } | kwargs

  sampler_kwargs, sampler_required = shared.get_default_signature(
      local_sampler)
  sampler_kwargs.setdefault("jit", True)
  sampler_kwargs.update(
      {k: defaults[k] for k in sampler_required if k in defaults})
  sampler_required = sampler_required - sampler_kwargs.keys()
  sampler_kwargs.update(
      {k: defaults[k] for k in sampler_kwargs if k in defaults})
  sampler_required = sampler_required - sampler_kwargs.keys()

  if sampler_required:
    raise ValueError(
        "Unexpected required arguments: "
        f"{','.join(sampler_required)}. Probably file a bug, but "
        "you can try to manually supply them as keywords."
    )
  return sampler_kwargs


def get_sampler_kwargs(sampler, n_features, kwargs):
  """Sets defaults and merges user-provided adaptation keywords."""
  #  We support `num_chains` everywhere else, so support it here.
  if "num_chains" in kwargs:
    kwargs["n_chains"] = kwargs["num_chains"]
  defaults = {
      "n_loop_training": 5,
      "n_loop_production": 5,
      "n_local_steps": 50,
      "n_global_steps": 50,
      "n_chains": 20,
      "n_epochs": 30,
      "learning_rate": 0.01,
      "max_samples": 10_000,
      "momentum": 0.9,
      "batch_size": 10_000,
      "use_global": True,
      "global_sampler": None,
      "logging": True,
      "keep_quantile": 0.,
      "local_autotune": None,
      "train_thinning": 1,
      "output_thinning": 1,
      "n_sample_max": 10_000,
      "precompile": False,
      "verbose": False,
      "n_dim": n_features,
      "data": {}} | kwargs
  sampler_kwargs, sampler_required = shared.get_default_signature(sampler)
  sampler_kwargs.update(
      {k: defaults[k] for k in sampler_required if k in defaults})
  sampler_required = (sampler_required -
                      {"nf_model", "local_sampler", "rng_key", "kwargs"})
  sampler_required = sampler_required - sampler_kwargs.keys()

  if sampler_required:
    raise ValueError(
        "Unexpected required arguments: "
        f"{','.join(sampler_required)}. Probably file a bug, but "
        "you can try to manually supply them as keywords."
    )
  return defaults | sampler_kwargs


class _FlowMCSampler(shared.Base):
  """Base class for flowmc samplers."""
  name: str = ""
  nf_model: str = ""
  local_sampler: str = ""

  def _get_aux(self):
    flat, unflatten = jax.flatten_util.ravel_pytree(self.test_point)

    @jax.vmap
    def flatten(pytree):
      return jax.flatten_util.ravel_pytree(pytree)[0]

    constrained_log_density = self.constrained_log_density()
    def log_density(x, _):
      return constrained_log_density(unflatten(x)).squeeze()

    return log_density, flatten, unflatten, flat.shape[0]

  def get_kwargs(self, **kwargs):
    nf_model = _NF_MODELS[self.nf_model]
    local_sampler = _LOCAL_SAMPLERS[self.local_sampler]
    log_density, flatten, unflatten, n_features = self._get_aux()

    nf_model_kwargs = get_nf_model_kwargs(nf_model, n_features, kwargs)
    local_sampler_kwargs = get_local_sampler_kwargs(
        local_sampler, log_density, n_features, kwargs)
    sampler = Sampler.Sampler
    sampler_kwargs = get_sampler_kwargs(sampler, n_features, kwargs)
    extra_parameters = {"flatten": flatten,
                        "unflatten": unflatten,
                        "num_chains": sampler_kwargs["n_chains"],
                        "return_pytree": kwargs.get("return_pytree", False)}

    return {nf_model: nf_model_kwargs,
            local_sampler: local_sampler_kwargs,
            sampler: sampler_kwargs,
            "extra_parameters": extra_parameters}

  def __call__(self, seed, **kwargs):
    kwargs = self.get_kwargs(**kwargs)
    extra_parameters = kwargs["extra_parameters"]
    num_chains = extra_parameters["num_chains"]
    init_key, nf_key, seed = jax.random.split(seed, 3)
    initial_state = self.get_initial_state(
        init_key, num_chains=num_chains)
    initial_state = extra_parameters["flatten"](initial_state)
    nf_model = _NF_MODELS[self.nf_model]
    local_sampler = _LOCAL_SAMPLERS[self.local_sampler]

    model = nf_model(key=nf_key, **kwargs[nf_model])
    local_sampler = local_sampler(**kwargs[local_sampler])
    sampler = Sampler.Sampler
    nf_sampler = sampler(
        rng_key=seed,
        local_sampler=local_sampler,
        nf_model=model,
        **kwargs[sampler])
    nf_sampler.sample(initial_state, {})
    chains, *_ = nf_sampler.get_sampler_state().values()

    unflatten = jax.vmap(jax.vmap(extra_parameters["unflatten"]))
    pytree = self.transform_fn(unflatten(chains))
    if extra_parameters["return_pytree"]:
      return pytree
    else:
      if hasattr(pytree, "_asdict"):
        pytree = pytree._asdict()
      elif not isinstance(pytree, dict):
        pytree = {"var0": pytree}
      return az.from_dict(posterior=pytree)


class RealNVPMALA(_FlowMCSampler):
  name = "flowmc_realnvp_mala"
  nf_model = "real_nvp"
  local_sampler = "mala"


class RealNVPHMC(_FlowMCSampler):
  name = "flowmc_realnvp_hmc"
  nf_model = "real_nvp"
  local_sampler = "hmc"


class MaskedCouplingRQSplineMALA(_FlowMCSampler):
  name = "flowmc_rqspline_mala"
  nf_model = "masked_coupling_rq_spline"
  local_sampler = "mala"


class MaskedCouplingRQSplineHMC(_FlowMCSampler):
  name = "flowmc_rqspline_hmc"
  nf_model = "masked_coupling_rq_spline"
  local_sampler = "hmc"



# Copyright 2024 The bayeux Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""API entry point."""

import dataclasses

from bayeux import mcmc
from bayeux import optimize
from bayeux import vi
from bayeux._src import shared
import jax
import oryx

_MODULES = (mcmc, optimize, vi)
_REQUIRED_KWARGS = (
    "log_density",
    "test_point",
    "transform_fn",
    "inverse_transform_fn",
    "inverse_log_det_jacobian",
    "initial_state",)


class _Namespace:

  def __init__(self):
    self.methods = []

  def __repr__(self):
    return "\n".join(self.methods)

  def __setclass__(self, clas, parent):
    kwargs = {k: getattr(parent, k) for k in _REQUIRED_KWARGS}
    setattr(self, clas.name, clas(**kwargs))
    self.methods.append(clas.name)


def is_tfp_bijector(bij):
  return hasattr(bij, "__class__") and bij.__class__.__module__.startswith(
      "tensorflow_probability")


@dataclasses.dataclass
class Model(shared.Base):
  """Base model class.

  Models consist of a possibly unnormalized log density function, a point that
  the function can act on, in that `log_density(test_point)` returns a scalar,
  and an optional `transform_fn`, which is a JAX function transforming
  a point from the reals into the support of the `log_density`. For example,
  if the `test_point` is a positive scalar, `transform_fn` may be `jnp.exp` or
  `jax.nn.softmax`.
  """
  _namespaces = None

  def _init(self):
    self._namespaces = []
    if is_tfp_bijector(self.transform_fn):
      self.transform_fn = self.transform_fn
      if self.inverse_transform_fn is None:
        self.inverse_transform_fn = self.transform_fn.inverse
      if self.inverse_log_det_jacobian is None:
        self.inverse_log_det_jacobian = (
            self.transform_fn.inverse_log_det_jacobian
        )
    if self.inverse_transform_fn is None:
      self.inverse_transform_fn = oryx.core.inverse(self.transform_fn)
    for module in _MODULES:
      k = _Namespace()
      module_name = module.__name__.split(".")[-1]
      for class_name in module.__all__:
        clas = getattr(module, class_name)
        k.__setclass__(clas, self)

      # guard against no optional libraries being installed.
      if module.__all__:
        setattr(self, module_name, k)
        self._namespaces.append(module_name)

  def __post_init__(self):
    self._init()
    super().__init__(log_density=self.log_density,
                     test_point=self.test_point,
                     transform_fn=self.transform_fn,
                     initial_state=self.initial_state,
                     inverse_transform_fn=self.inverse_transform_fn,
                     inverse_log_det_jacobian=self.inverse_log_det_jacobian)

  def __repr__(self):
    methods = []
    for key, values in self.methods.items():
      methods.append(key)
      methods.append("\t." + "\n\t.".join(values))
    return "\n".join(methods)

  @property
  def methods(self):
    methods = {}
    for name in self._namespaces:
      methods[name] = getattr(self, name).methods
    return methods

  @classmethod
  def from_tfp(cls, pinned_joint_distribution, initial_state=None):
    log_density = pinned_joint_distribution.log_prob
    test_point = pinned_joint_distribution.sample_unpinned(
        seed=jax.random.PRNGKey(0))
    transform_fn = (
        pinned_joint_distribution.experimental_default_event_space_bijector()
    )
    inverse_transform_fn = transform_fn.inverse
    inverse_log_det_jacobian = transform_fn.inverse_log_det_jacobian
    return cls(
        log_density=log_density,
        test_point=test_point,
        transform_fn=transform_fn,
        initial_state=initial_state,
        inverse_transform_fn=inverse_transform_fn,
        inverse_log_det_jacobian=inverse_log_det_jacobian)

  @classmethod
  def from_numpyro(cls, numpyro_fn, initial_state=None):
    import numpyro  # pylint: disable=g-import-not-at-top

    def log_density(*args, **kwargs):
      # This clause is only required because the tfp vi routine tries to
      # pass dictionaries as keyword arguments, so this allows either
      # log_density(params) or log_density(**params)
      if args:
        x = args[0]
      else:
        x = kwargs
      return numpyro.infer.util.log_density(numpyro_fn, (), {}, x)[0]

    test_point = numpyro.infer.Predictive(
        numpyro_fn, num_samples=1)(jax.random.PRNGKey(0))
    test_point = {k: v[0] for k, v in test_point.items() if k != "observed"}

    def transform_fn(x):
      return numpyro.infer.util.constrain_fn(numpyro_fn, (), {}, x)

    return cls(
        log_density=log_density,
        test_point=test_point,
        transform_fn=transform_fn,
        initial_state=initial_state)

  @classmethod
  def from_pymc(cls, pm_model, initial_state=None):
    import pymc as pm  # pylint: disable=g-import-not-at-top
    import pymc.sampling.jax as pm_jax  # pylint: disable=g-import-not-at-top

    class Inverse(pm.logprob.transforms.Transform):
      def __init__(self, transform):
        self._transform = transform

      def forward(self, value, *inputs):
        """Apply the transformation."""
        return self._transform.backward(value, *inputs)

      def backward(self, value, *inputs):
        return self._transform.forward(value, *inputs)

    uc_model = pm.model.transform.conditioning.remove_value_transforms(pm_model)
    logp = pm_jax.get_jaxified_logp(uc_model)

    rvs_to_inverse = {
        k: None if v is None else Inverse(v)
        for k, v in pm_model.rvs_to_transforms.items()}
    rvs = pm_model.free_RVs
    inv_rvs = pm.logprob.utils.replace_rvs_by_values(
        rvs,
        rvs_to_values=pm_model.rvs_to_values,
        rvs_to_transforms=rvs_to_inverse)
    values = pm_model.value_vars
    names = [v.name for v in rvs]

    fwd = pm_jax.get_jaxified_graph(
        inputs=values,
        outputs=pm_model.replace_rvs_by_values(rvs))

    bwd = pm_jax.get_jaxified_graph(
        inputs=values,
        outputs=pm_model.replace_rvs_by_values(inv_rvs))
    def logp_wrap(*args, **kwargs):
      # This clause is only required because the tfp vi routine tries to
      # pass dictionaries as keyword arguments, so this allows either
      # log_density(params) or log_density(**params)
      if args:
        kwargs = args[0]
      return logp([kwargs[k] for k in names])

    def fwd_wrap(args):
      ret = fwd(*[args[k] for k in names])
      return dict(zip(names, ret))

    def bwd_wrap(args):
      ret = bwd(*[args[k] for k in names])
      return dict(zip(names, ret))

    test_point = uc_model.initial_point()
    return cls(
        log_density=logp_wrap,
        test_point=test_point,
        transform_fn=fwd_wrap,
        inverse_transform_fn=bwd_wrap,
        initial_state=initial_state)



# Copyright 2024 The bayeux Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Shared functionality for MCMC sampling."""

import dataclasses
import functools
import inspect
from typing import Any, Callable, Optional

from bayeux._src import debug
from bayeux._src import initialization
from bayeux._src import types
import jax
import jax.numpy as jnp
import oryx


def map_fn(chain_method, fn):
  if chain_method == "parallel":
    return jax.pmap(fn)
  elif chain_method == "vectorized":
    return jax.vmap(fn)
  elif chain_method == "sequential":
    return functools.partial(jax.tree.map, fn)
  raise ValueError(f"Chain method {chain_method} not supported.")


def update_with_kwargs(
    defaults: dict[str, Any],
    *,
    reqd: Optional[set[str]] = None,
    kwargs: dict[str, Any],
):
  """Updates a defaults dictionary, overwriting keys, but not adding new ones.

  This updates `defaults` in-place.

  Args:
    defaults: Dictionary of arguments.
    reqd: Set of allowed new keys.
    kwargs: Dictionary of key/values to overwrite in default.

  Returns:
    The defaults dictionary. Keys will be the same as the existing ones, or from
    the set `reqd`. Note that this is also updated in-place.
  """
  if reqd is None:
    reqd = set()
  defaults.update(
      (k, kwargs[k]) for k in (defaults.keys() | reqd) & kwargs.keys())
  return defaults


def _default_init(
    *,
    initial_state,
    test_point,
    inverse_transform_fn,
    transform_fn,
    num_points,
    key
):
  """Initialization in case there is no explicit init provided."""
  if initial_state is None:
    return initialization.uniform(
        test_point=test_point,
        inverse_transform_fn=inverse_transform_fn,
        transform_fn=transform_fn,
        num_points=num_points,
        key=key)
  else:
    return initial_state


def constrain(
    transform_fn: types.JAXFn,
    inverse_log_det_jacobian: Optional[types.JAXFn] = None,
) -> Callable[[types.LogDensityFn], types.LogDensityFn]:
  """Returns a log density function that operates in an unconstrained space.

  Adapted from oryx (https://github.com/jax-ml/oryx)

  Args:
    transform_fn: Constraining bijector, mapping from R^n to the support of the
      target log density.
    inverse_log_det_jacobian: Optional inverse log det jacobian, if known.
  """
  if inverse_log_det_jacobian is None:
    inverse_log_det_jacobian = oryx.core.ildj(transform_fn)

  def wrap_log_density(target_log_density):
    def wrapped(args):
      mapped_args = transform_fn(args)
      ildjs = inverse_log_det_jacobian(mapped_args)
      return target_log_density(mapped_args) - jnp.sum(
          jnp.array(jax.tree_util.tree_leaves(ildjs)))

    return wrapped

  return wrap_log_density


def get_default_signature(fn):
  defaults = {}
  required = set()
  for key, val in inspect.signature(fn).parameters.items():
    if val.default is inspect.Signature.empty:
      required.add(key)
    else:
      defaults[key] = val.default
  return defaults, required


def _nothing(x: types.Point) -> types.Point:
  return x


@dataclasses.dataclass
class Base:
  """Base class for MCMC sampling."""
  log_density: types.LogDensityFn
  test_point: types.Point
  transform_fn: types.JAXFn = _nothing
  inverse_transform_fn: Optional[types.JAXFn] = None
  inverse_log_det_jacobian: Optional[types.JAXFn] = None

  initial_state: Optional[types.Point] = None

  def get_initial_state(self, key, num_chains=8):
    return _default_init(
        initial_state=self.initial_state,
        test_point=self.test_point,
        inverse_transform_fn=self.inverse_transform_fn,
        transform_fn=self.transform_fn,
        num_points=num_chains,
        key=key)

  def get_kwargs(self, **kwargs):
    raise NotImplementedError()

  def constrained_log_density(self):
    return constrain(
        self.transform_fn,
        inverse_log_det_jacobian=self.inverse_log_det_jacobian,
    )(self.log_density)

  def debug(
      self,
      seed,
      verbosity=2,
      catch_exceptions: bool = True,
      printer=print,
      kwargs=None,
  ):
    return debug.debug(
        self,
        seed,
        verbosity=verbosity,
        catch_exceptions=catch_exceptions,
        printer=printer,
        kwargs=kwargs,
    )



# Copyright 2024 The bayeux Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""optax specific code."""
import functools

from bayeux._src.optimize import shared
import jax
import optax
import optax.contrib


class _OptaxOptimizer(shared.Optimizer):
  """Base class for optax optimizers."""
  _base = optax

  def get_kwargs(self, **kwargs):
    kwargs = self.default_kwargs() | kwargs
    optimizer = getattr(self._base, self.optimizer)
    return {optimizer: shared.get_optimizer_kwargs(optimizer, kwargs),
            "extra_parameters": shared.get_extra_kwargs(kwargs)}

  def __call__(self, seed, **kwargs):
    kwargs = self.get_kwargs(**kwargs)
    fun, initial_state, apply_transform = self._prep_args(seed, kwargs)

    optimizer_fn = getattr(self._base, self.optimizer)
    optimizer = optimizer_fn(**kwargs[optimizer_fn])
    num_iters = kwargs["extra_parameters"]["num_iters"]
    optimizer = functools.partial(
        _optax_fit,
        neg_log_density=fun,
        optimizer=optimizer,
        num_iters=num_iters)
    chain_method = kwargs["extra_parameters"]["chain_method"]
    mapped_optimizer = self._map_optimizer(chain_method, optimizer)
    res, loss = mapped_optimizer(initial_state)
    if apply_transform:
      return shared.OptimizerResults(
          params=self.transform_fn(res[0]), state=res[1][0], loss=loss)
    else:
      return shared.OptimizerResults(*res, loss=loss)


class AdaBelief(_OptaxOptimizer):
  name = "optax_adabelief"
  optimizer = "adabelief"


class Adafactor(_OptaxOptimizer):
  name = "optax_adafactor"
  optimizer = "adafactor"

  def default_kwargs(self) -> dict[str, float]:
    kwargs = super().default_kwargs()
    kwargs["decay_rate"] = 0.90
    kwargs["learning_rate"] = 1e-4
    return kwargs


class Adagrad(_OptaxOptimizer):
  name = "optax_adagrad"
  optimizer = "adagrad"


class Adam(_OptaxOptimizer):
  name = "optax_adam"
  optimizer = "adam"


class Adamw(_OptaxOptimizer):
  name = "optax_adamw"
  optimizer = "adamw"


class Adamax(_OptaxOptimizer):
  name = "optax_adamax"
  optimizer = "adamax"


class Adamaxw(_OptaxOptimizer):
  name = "optax_adamaxw"
  optimizer = "adamaxw"


class Amsgrad(_OptaxOptimizer):
  name = "optax_amsgrad"
  optimizer = "amsgrad"


class Fromage(_OptaxOptimizer):
  name = "optax_fromage"
  optimizer = "fromage"

  def default_kwargs(self) -> dict[str, float]:
    kwargs = super().default_kwargs()
    kwargs["learning_rate"] = 1e-4
    return kwargs


class Lamb(_OptaxOptimizer):
  name = "optax_lamb"
  optimizer = "lamb"

  def default_kwargs(self) -> dict[str, float]:
    kwargs = super().default_kwargs()
    kwargs["learning_rate"] = 1e-4
    return kwargs

# Too finicky
# class Lars(_OptaxOptimizer):
#   name = "optax_lars"
#   optimizer = "lars"


class Lion(_OptaxOptimizer):
  name = "optax_lion"
  optimizer = "lion"

  def default_kwargs(self) -> dict[str, float]:
    kwargs = super().default_kwargs()
    kwargs["learning_rate"] = 1e-4
    return kwargs


class NoisySgd(_OptaxOptimizer):
  name = "optax_noisy_sgd"
  optimizer = "noisy_sgd"

  def default_kwargs(self) -> dict[str, float]:
    kwargs = super().default_kwargs()
    kwargs["learning_rate"] = 1e-4
    return kwargs


class Novograd(_OptaxOptimizer):
  name = "optax_novograd"
  optimizer = "novograd"

  def default_kwargs(self) -> dict[str, float]:
    kwargs = super().default_kwargs()
    kwargs["learning_rate"] = 1e-4
    return kwargs

# Cannot get this to pass tests.
# class OptimisticGradientDescent(_OptaxOptimizer):
#   name = "optax_optimistic_gradient_descent"
#   optimizer = "optimistic_gradient_descent"


# Re-enable this when we figure out how to pass a seed to an optimizer.
# class Dpsgd(_OptaxOptimizer):
#   name = "optax_dpsgd"
#   optimizer = "dpsgd"


class Radam(_OptaxOptimizer):
  name = "optax_radam"
  optimizer = "radam"

  def default_kwargs(self) -> dict[str, float]:
    kwargs = super().default_kwargs()
    kwargs["learning_rate"] = 1e-4
    return kwargs


class Rmsprop(_OptaxOptimizer):
  name = "optax_rmsprop"
  optimizer = "rmsprop"

  def default_kwargs(self) -> dict[str, float]:
    kwargs = super().default_kwargs()
    kwargs["learning_rate"] = 1e-4
    return kwargs


class ScheduleFree(_OptaxOptimizer):
  _base = optax.contrib
  name = "optax_schedule_free"
  optimizer = "schedule_free"

  def default_kwargs(self) -> dict[str, float]:
    kwargs = super().default_kwargs()
    base_optimizer = optax.adam(
        **shared.get_optimizer_kwargs(optax.adam, kwargs))
    kwargs["base_optimizer"] = base_optimizer
    return kwargs


class Sgd(_OptaxOptimizer):
  name = "optax_sgd"
  optimizer = "sgd"

  def default_kwargs(self) -> dict[str, float]:
    kwargs = super().default_kwargs()
    kwargs["learning_rate"] = 1e-4
    return kwargs


class Sm3(_OptaxOptimizer):
  name = "optax_sm3"
  optimizer = "sm3"


class Yogi(_OptaxOptimizer):
  name = "optax_yogi"
  optimizer = "yogi"


def _optax_fit(params, neg_log_density, optimizer, num_iters):
  opt_state = optimizer.init(params)

  def step(params_and_state, _):
    params, opt_state = params_and_state
    loss, grads = jax.value_and_grad(neg_log_density)(params)
    updates, opt_state = optimizer.update(grads, opt_state, params)
    params = optax.apply_updates(params, updates)
    return (params, opt_state), loss

  init = (params, opt_state)
  return jax.lax.scan(step, init, xs=None, length=num_iters)
